;;; cgreek-tex.el --- Ibycus4 interface for classical Greek

;; Copyright (C) 2003, 2005, 2010
;; National Institute of Advanced Industrial Science and Technology (AIST)
;; Registration Number: H15PRO 109

;; Author: TAKAHASHI Naoto <ntakahas@m17n.org>
;; Keywords: multilingual, classical Greek, Unicode, LaTeX, Ibycus4

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth
;; Floor, Boston, MA 02110-1301, USA.

;;; Code:

(require 'robin)

(defvar cgreek-begingreek "\\begin{greek}"
  "*String indicating the beginning of Greek text")

(defvar cgreek-endgreek "\\end{greek}"
  "*String indicating the end of Greek text")

(defconst cgreek-category-table (make-category-table))
(define-category ?l "has latex2e transliteration" cgreek-category-table)
(define-category ?i "has ibycus4 transliteration" cgreek-category-table)

;; Latin-1 Supplement, Latin Extended-A, Latin-Extended-B
(let ((i #x00a0))
  (while (< i #x017f)
    (modify-category-entry i ?l cgreek-category-table)
    (setq i (1+ i))))

;; Latin Extended-B
(let ((l '(#x01cd #x01ce #x01cf #x01d0 #x01d1 #x01d2 #x01d3 #x01d4
	   #x01e2 #x01e3 #x01e6 #x01e7 #x01e8 #x01e9 #x01f0
	   #x01f4 #x01f5 #x01f8 #x01f9 #x01fc #x01fd #x01fe #x01ff
	   #x021e #x021f #x0226 #x0227 #x0228 #x0229 #x022e #x022f
	   #x0232 #x0233)))
  (while l
    (modify-category-entry (car l) ?l cgreek-category-table)
    (setq l (cdr l))))

;; Latin Extended Additional
(let ((i #x1e00))
  (while (<= i #x1eff)
    (modify-category-entry i ?l cgreek-category-table)
    (setq i (1+ i))))

;; Greek
(let ((i #x037a))
  (while (<= i #x03ce)
    (modify-category-entry i ?i cgreek-category-table)
    (setq i (1+ i))))

;; punctuations (only in Greek), archaic letters
(let ((l '(#x00ab #x00bb #x2039 #x203a
	   #x0342 #x0343 #x0344 #x0345
	   #x03d8 #x03d9 #x03dc #x03dd #x03e1 #x03f2 #x03f9)))
  (while l
    (modify-category-entry (car l) ?i cgreek-category-table)
    (setq l (cdr l))))

;; Greek Extended
(let ((i #x1f00))
  (while (<= i #x1fff)
    (modify-category-entry i ?i cgreek-category-table)
    (setq i (1+ i))))

;; punctuations (available both in Greek and Latin environments)
(let ((l '(#x2010 #x2014 #x2016 #x2018 #x2019 #x201c #x201d #x2020 #x2021
	   #x2153 #x2154 #x2159 #x215b
	   #x2190 #x2192 #x2212))
      c)
  (while l
    (setq c (car l))
    (modify-category-entry c ?l cgreek-category-table)
    (modify-category-entry c ?i cgreek-category-table)
    (setq l (cdr l))))

;;; Unicode to TeX

(defun cgreek-unicode-to-tex-region (begin end)
  "Convert Unicode characters in the regin into TeX notations."
  (interactive "*r")
  (save-restriction
    (narrow-to-region begin end)
    (cgreek-unicode-to-tex-buffer)))

(defun cgreek-unicode-to-tex-buffer ()
  "Convert Unicode characters in the current buffer into TeX notations."
  (interactive "*")
  (with-category-table cgreek-category-table

    ;; Unicode (Latin) to LaTeX2e
    (goto-char (point-min))
    (while (re-search-forward "\\cl" nil t)
      (forward-char -1)
      (insert
       (get-char-code-property (following-char) 'latin-latex2e))
      (delete-char 1))

    ;; Unicode (Greek) to Ibycus4

    ;; A quick hack for below dot
    (goto-char (point-min))
    (while (re-search-forward (concat "\\(\\ci\\)" (char-to-string #x0323))
			      nil t)
      (replace-match "\\\\d{\\1}"))

    ;; regexp matcher overflows for a long Greek sequence, so we first
    ;; divide Greek text into lines, then remove redundant pairs of
    ;; cgreek-endgreek and cgreek-begingreek.

    (goto-char (point-min))
    (while (re-search-forward "\\ci" nil t)
      (forward-char -1)
      (looking-at "\\(\\ci\\|[ ,.]\\)*\\ci")
      (save-restriction
	(narrow-to-region (match-beginning 0) (match-end 0))
	(insert cgreek-begingreek)
	(while (not (eobp))
	  (if (looking-at "\\Ci")
	      (forward-char 1)
	    (insert
	     (or (get-char-code-property (following-char) 'greek-ibycus4)
		 32))			; no transliteration available
	    (delete-char 1)))
	(insert cgreek-endgreek)))

    ;; remove redundant pairs of cgreek-endgreek and cgreek-begingreek
    (goto-char (point-min))
    (let ((redundant
	   (concat (regexp-quote cgreek-endgreek)
		   "\\([ ,.\n]*\\)"
		   (regexp-quote cgreek-begingreek))))
      (while (re-search-forward redundant nil t)
	(replace-match "\\1")))))

(defun cgreek-write-file-tex (filename)
  "Write current buffer into file FILENAME in the TeX format.
Makes buffer visit that file, and marks it not modified.
If the buffer is already visiting a file, you can specify
a directory name as FILENAME, to write a file of the same
old name in that directory."
  (interactive
   (list (if buffer-file-truename
	     (read-file-name "Write TeX file: " buffer-file-truename)
	   (read-file-name "Write TeX file: "
			   (cdr (assq 'default-directory
				      (buffer-local-variables)))
			   nil nil (buffer-name)))))
  (let ((old-buffer (current-buffer)))
    (with-temp-buffer
      (insert-buffer-substring old-buffer)
      (cgreek-unicode-to-tex-buffer)
      (write-file (expand-file-name filename)))
    (set-buffer-modified-p nil)))

;;; TeX to Unicode

(defun cgreek-tex-to-unicode-region (begin end)
  "Convert TeX notations in the region into Unicode characters.
Regions surrounded by the value of cgreek-begingreek and
cgreek-endgreek are interpreted as Ibycus4 notations."
  (interactive "*r")
  (save-restriction
    (narrow-to-region begin end)
    (cgreek-tex-to-unicode-buffer)))

(defun cgreek-tex-to-unicode-buffer nil
  "Convert TeX notations in the current buffer into Unicode
characters.  Regions surrounded by cgreek-begingreek cgreek-endgreek
are interpreted as Ibycus4 notations."
  (interactive "*")

  (goto-char (point-min))
  (while (search-forward cgreek-begingreek nil t)
    (replace-match "")
    (robin-convert-region
     (point)
     (progn
       (if (search-forward cgreek-endgreek nil 0)
	   (replace-match ""))
       (point))
     "greek-ibycus4"))

  ;; A quick hack for below dot
  (goto-char (point-min))
  (while (re-search-forward "\\\\d{\\(.\\)}" nil t)
    (replace-match "\\1")
    (insert #x0323))

  (robin-convert-region (point-min) (point-max) "latin-latex2e"))

(defun cgreek-find-file-tex (filename)
  "Edit file FILENAME encoded in the TeX format.
Switch to a buffer visiting file FILENAME,
creating one if none already exists."
  (interactive "FFind TeX file: ")
  (find-file filename)
  (cgreek-tex-to-unicode-buffer)
  (set-buffer-modified-p nil)
  (goto-char (point-min)))

(provide 'cgreek-tex)

;;; cgreek-tex.el ends here
