;;; cgreek-tlg.el --- utilities for reading TLG CD-ROM

;; Copyright (C) 2003, 2005, 2008, 2010
;; National Institute of Advanced Industrial Science and Technology (AIST)
;; Registration Number: H15PRO 109

;; Author: TAKAHASHI Naoto <ntakahas@m17n.org>
;; Keywords: multilingual, classical Greek, Unicode, TLG

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.
;; the Free Software Foundation, Inc., 51 Franklin Street, Fifth
;; Floor, Boston, MA 02110-1301, USA.

;;; Code

(defvar cgreek-tlg-directory "/media/cdrom"
  "*Directory in which TLG files reside.")

(defvar cgreek-tlg-cdrom-version "E"
  "*Set this variable to \"D\" to avoid the error in Disk D.")

(defvar cgreek-tlg-my-authtab-file nil
  "*Absolute pathname to a file containing a customised author list.
Each line in the file begins with the string \"TLG\" directly followed
by four digits.  Rest of the line can be anything.")

(defvar cgreek-tlg2emacs-program
  (expand-file-name
   "tlg2emacs" (file-name-directory (locate-library "cgreek")))
  "*Program name for invoking a TLG-to-CGreek converter.")

(if (null (assoc cgreek-tlg2emacs-program process-coding-system-alist))
    (setq process-coding-system-alist
	  (cons (cons cgreek-tlg2emacs-program
		      (cons 'emacs-mule 'no-conversion))
		process-coding-system-alist)))

(defvar cgreek-tlg2citation-program
  (expand-file-name
   "tlg2cit" (file-name-directory (locate-library "cgreek")))
  "*Program name to extract citation information from tlg txt file.")

(if (null (assoc cgreek-tlg2citation-program process-coding-system-alist))
    (setq process-coding-system-alist
	  (cons (cons cgreek-tlg2citation-program
		      (cons 'no-conversion 'no-conversion))
		process-coding-system-alist)))

(defconst cgreek-tlg-end-of-file 0
  "Type code for end-of-file in TLG IDT file.")

(defconst cgreek-tlg-new-author 1
  "Type code for `new author' in TLG IDT file.")

(defconst cgreek-tlg-new-work 2
  "Type code for `new work' in TLG IDT file.")

(defconst cgreek-tlg-author-work 16
  "Type code for `description of ID fields a..b' in TLG IDT file.
If followed by numeric value 0, means ID level a; if 1, means b.")

(defconst cgreek-tlg-citation 17
  "Type code for `description of ID fields n, v..z' in TLG IDT file.
In case of document, followed by numeric value 0 and means ID level n.
Otherwise, followed by 4..0 meaning v..z, respectively.")

(defconst cgreek-tlg-citation-regexp
  "\\([^\t]*\\)\t\\([^\t]*\\)\t\\([^\t]*\\)\t\\([^\t]*\\)\t\\([^\t]*\\)\t\\([^\t]*\\)\t"
  "Regular expression used to get citation ID value in a tlg file.
There are 6 strings separated by tab, of which
(match-string 1) corresponds to level n citation,
(match-string 2) corresponds to level v citation,
(match-string 3) corresponds to level w citation,
(match-string 4) corresponds to level x citation,
(match-string 5) corresponds to level y citation and
(match-string 6) corresponds to level z citation.")

(defvar cgreek-tlg-authtab-buffer nil
  "Buffer used for selecting an author.")

(defvar cgreek-tlg-temporary-file
  (make-temp-file "cgreek-tlg")
  "Temporary file for citation extraction.")

(defvar cgreek-tlg-temporary-buffer
  (get-buffer-create " cgreek-tlg-temporary-buffer")
  "Temporary file for beta code conversion.")

(defvar cgreek-tlg-author-name nil
  "Athor name of the work displayed in the current buffer.")

(make-variable-buffer-local 'cgreek-tlg-author-name)

(defvar cgreek-tlg-work-name nil
  "Work name displayed in the current buffer.")

(make-variable-buffer-local 'cgreek-tlg-work-name)

(defvar cgreek-tlg-citation-buffer nil
  "Buffer containing the citation information of the current buffer.")

(make-variable-buffer-local 'cgreek-tlg-citation-buffer)

(defvar cgreek-tlg-citation-description nil
  "Citation descriptions of the current buffer.
Plist of the form (level-1 desc-1 level-2 desc-2 ...) where
level-x is integer from 2 (level v) to 6 (level z or n) and
desc-x is string like \"line\", \"section\", etc.")

(make-variable-buffer-local 'cgreek-tlg-citation-description)

(defvar cgreek-tlg-idt-buffer nil
  "Buffer containing the work list of the same author.")

(make-variable-buffer-local 'cgreek-tlg-idt-buffer)

(defun cgreek-tlg-parse-authtab ()
  "Parse and display the \"authtab.dir\" file.
The parsed result is shown in the *TLG authtab* buffer.
Hitting RET opens the file for that author."
  (interactive)
  (cond

   ;; already created
   ((buffer-live-p cgreek-tlg-authtab-buffer)
    (switch-to-buffer cgreek-tlg-authtab-buffer))

   ;; customised file is set
   (cgreek-tlg-my-authtab-file
    (find-file cgreek-tlg-my-authtab-file)

    ;; a little make-up
    (goto-char (point-min))
    (while (looking-at "TLG[0-9][0-9][0-9][0-9]")
      (forward-char 7)
      (if (looking-at "[ \t]+")
	  (goto-char (match-end 0)))
      (add-text-properties
       (point)
       (progn (end-of-line) (point))
       '(mouse-face highlight))
      (beginning-of-line 2))		; go to next line's beginning

    ;; local bindings
    (local-set-key "q" '(lambda () (interactive) (kill-buffer nil)))
    (local-set-key [mouse-2] 'cgreek-tlg-parse-idt-mouse)
    (local-set-key "\r" 'cgreek-tlg-parse-idt-keyboard)
    (local-set-key "\C-m" 'cgreek-tlg-parse-idt-keyboard)

    (goto-char (point-min))
    (set-buffer-modified-p nil)
    (setq cgreek-tlg-authtab-buffer (current-buffer)))

   ;; newly create
   (t
    (let ((raw (generate-new-buffer "authtab.dir"))
	  (cooked (get-buffer-create "*TLG authtab*"))
	  (str nil))
      (message "Reading authtab ... ")

      ;; initialise temporary buffer
      (set-buffer cgreek-tlg-temporary-buffer)
      (erase-buffer)

      ;; initialise cooked (user visible) authtab
      (set-buffer cooked)
      (setq buffer-read-only nil)
      (erase-buffer)

      ;; prepare the original file contents
      (set-buffer raw)
      (set-buffer-multibyte nil)
      (insert-file-contents-literally
       (expand-file-name "authtab.dir" cgreek-tlg-directory))
      (re-search-forward ".TLG....[^\xff]+\xff") ; skip header

      (while (not (looking-at "\*END[\0\xff]+"))
	(cond

	 ;; plain text
	 ((or (looking-at "TLG[0-9][0-9][0-9][0-9] ")
	      (looking-at "DOCCAN[0-9]+ "))
	  (setq str (buffer-substring (match-beginning 0) (match-end 0)))
	  (goto-char (match-end 0)))

	 ;; Beta codes go to cgreek-tlg-temporary-buffer,
	 ;; separated by \0, for later conversion.
	 ;; For the mean time \0 is inserted in cooked as position marker.
	 ((looking-at "[\x20-\x7e]+")
	  (save-excursion
	    (set-buffer cgreek-tlg-temporary-buffer)
	    (insert-buffer-substring raw (match-beginning 0) (match-end 0))
	    (insert 0))
	  (setq str (char-to-string 0))
	  (goto-char (match-end 0)))

	 ;; special codes
	 ((= (following-char) ?\x80)
	  (setq str ", AKA ")
	  (forward-char 1))

	 ((= (following-char) ?\x81)
	  (setq str ". Remarks: ")
	  (forward-char 2))

	 ((= (following-char) ?\x82)
	  (setq str ". Size: ")
	  (forward-char 2))

	 ((= (following-char) ?\x83)
	  (setq str
		(concat ". Language: "
			(char-to-string (char-after (1+ (point))))))
	  (forward-char 2))

	 ;; field separator
	 ((looking-at "[\xff]+")
	  (setq str "\n")
	  (goto-char (match-end 0)))

	 (t
	  (error "Unexpected code %c at %d" (following-char) (point))))

	;; insert cooked string
	(save-excursion
	  (set-buffer cooked)
	  (insert str)))

      ;; doccan files are not supported yet
      (let ((kill-whole-line t))
	(goto-char (point-min))
	(while (re-search-forward "^DOCCAN[0-9]+ " nil t)
	  (beginning-of-line)
	  (kill-line)))

      ;; convert beta code in cgreek-tlg-temporary-buffer
      ;; note that \0 becomes \n
      (set-buffer cgreek-tlg-temporary-buffer)
      (call-process-region
       (point-min) (point-max)
       cgreek-tlg2emacs-program t t)
      (goto-char (point-min))

      ;; replace each \0 in cooked with a line in cgreek-tlg-temporary-buffer
      (set-buffer cooked)
      (goto-char (point-min))
      (let (start end (zero (char-to-string 0)))
	(while (search-forward zero nil t)
	  (save-excursion
	    (set-buffer cgreek-tlg-temporary-buffer)
	    (setq start (point))
	    (end-of-line)
	    (setq end (point))
	    (forward-char 1))		; go to next line's beginning
	  (replace-match "")
	  (add-text-properties
	   (point)
	   (progn
	     (insert-buffer-substring cgreek-tlg-temporary-buffer start end)
	     (point))
	   '(mouse-face highlight))))

      (goto-char (point-min))

      ;; local bindings
      (local-set-key "q" '(lambda () (interactive) (kill-buffer nil)))
      (local-set-key [mouse-2] 'cgreek-tlg-parse-idt-mouse)
      (local-set-key "\r" 'cgreek-tlg-parse-idt-keyboard)
      (local-set-key "\C-m" 'cgreek-tlg-parse-idt-keyboard)
      (set-buffer-modified-p nil)
      (setq buffer-read-only t)

      ;; clean up
      (kill-buffer raw)
      (switch-to-buffer cooked)
      (setq cgreek-tlg-authtab-buffer (current-buffer))
      (message nil)))))

(defun cgreek-tlg-parse-idt-mouse (click)
  "Wrapper of cgreek-tlg-parse-idt for interactive use.
In the \"*TLG authtab*\" buffer, open line-initial tlg???? file."
  (interactive "e")
  (let* ((start (event-start click))
	 (window (car start))
	 (pos (car (cdr start))))
    (select-window window)
    (goto-char pos))
  (cgreek-tlg-parse-idt-keyboard))

(defun cgreek-tlg-parse-idt-keyboard ()
  "Wrapper of cgreek-tlg-parse-idt for interactive use.
In the \"*TLG authtab*\" buffer, open line-initial tlg???? file."
  (interactive)

  (cond
   ((eq (current-buffer) cgreek-tlg-authtab-buffer)
    (beginning-of-line)
    (if (not (looking-at "TLG[0-9][0-9][0-9][0-9]"))
	(message "N/A for this file")
      (switch-to-buffer
       (cgreek-tlg-parse-idt
	(concat
	 (downcase (buffer-substring (match-beginning 0) (match-end 0)))
	 ".idt")))))

   (cgreek-tlg-idt-buffer
    (switch-to-buffer
     (cgreek-tlg-parse-idt
      (substring cgreek-tlg-idt-buffer
		 1 (1- (length cgreek-tlg-idt-buffer))))))

   (t					; should never happen
    (error "cgreek-tlg-parse-idt-keyboard called from an unexpected buffer"))))

(defun cgreek-tlg-parse-idt (filename)
  "Parse the TLG IDT file FILENAME.
The result is stored in a buffer, which will be the return value."
  (let ((raw nil)
	(cooked (get-buffer-create (concat "*" filename "*"))))

    (set-buffer cooked)
    (if (/= (point-min) (point-max))

	;; OK, already created.  Just return it.
	cooked

      ;; not created yet.  Make it.
      (message "Reading %s ..." filename)

      ;; probability of "cooked being empty but read-only" is not zero
      (setq buffer-read-only nil)

      ;; initialise temporary buffer
      (set-buffer cgreek-tlg-temporary-buffer)
      (erase-buffer)

      ;; prepare the original file contents
      (setq raw (generate-new-buffer filename))
      (set-buffer raw)
      (set-buffer-multibyte nil)
      (insert-file-contents-literally
       (expand-file-name filename cgreek-tlg-directory))

      ;; extract citation description from FILENAME

      ;; Beta code goes to cgreek-tlg-temporary-buffer,
      ;; separated by \0, for later conversion.
      ;; For the mean time \0 is inserted in cooked as position marker
      (while (eq (char-after (point)) cgreek-tlg-new-author)
	(cgreek-tlg-parse-author cooked))
      (if (/= (char-after (point)) cgreek-tlg-end-of-file)
	  (error "Unexpected type code %s at %d"
		 (buffer-file-name)
		 (point)))

      ;; convert beta code in cgreek-tlg-temporary-buffer
      ;; note that \0 becomes \n
      (set-buffer cgreek-tlg-temporary-buffer)
      (call-process-region
       (point-min) (point-max)
       cgreek-tlg2emacs-program t t nil)
      (goto-char (point-min))

      ;; replace each \0 in cooked with a line in cgreek-tlg-temporary-buffer
      ;; do not forget to inherit text properties
      (set-buffer cooked)
      (goto-char (point-min))
      (let (start end prop)
	(while (search-forward (char-to-string 0) nil t)
	  (setq prop (text-properties-at (match-beginning 0)))
	  (replace-match "")
	  (save-excursion
	    (set-buffer cgreek-tlg-temporary-buffer)
	    (setq start (point))
	    (end-of-line)
	    (setq end (point))
	    (forward-char 1))		; go to next line's beginning
	  (add-text-properties
	   (point)
	   (progn
	     (insert-buffer-substring cgreek-tlg-temporary-buffer start end)
	     (point))
	   prop)))

      ;; local bindings
      (local-set-key "q" '(lambda () (interactive) (kill-buffer nil)))
      (local-set-key "a" 'cgreek-tlg-parse-authtab)
      (local-set-key [mouse-2] 'cgreek-tlg-parse-txt-mouse)
      (local-set-key "\r" 'cgreek-tlg-parse-txt-keyboard)
      (local-set-key "\C-m" 'cgreek-tlg-parse-txt-keyboard)
      (goto-char (point-min))
      (set-buffer-modified-p nil)
      (setq buffer-read-only t)

      ;; clean up
      (kill-buffer raw)
      (message nil)

      cooked)))

(defun cgreek-tlg-parse-author (cooked)
  "Process an author entry in TLG IDT file."

  (let (id desc next-author)

    ;; should be looking at:
    ;; ?\x01		type code 1 marks new author
    ;; [2 bytes]	length of the author section in bytes
    ;; [2 bytes]	author location in blocks
    ;; ?\xef		left nibble = 1110 : escape code
    ;;			right nibble = 1111 : ASCII string
    ;; ?\x80		escape level = 0 (author)
    ;; [n bytes]	ID value in string (with MSB on)
    ;; ?\xff		ASCII string terminator

    ;; just make sure
    (if (or (/= (char-after (+ (point) 5)) ?\xef)
	    (/= (char-after (+ (point) 6)) ?\x80))
      (error "Unexpected author ID value introducer %02x%02x"
	     (char-after (+ (point) 5))
	     (char-after (+ (point) 6))))

    (setq next-author
	  (+ (point)
	     (* 256 (char-after (1+ (point))))
	     (char-after (+ 2 (point)))))
    (forward-char 7)
    (setq id (cgreek-tlg-read-string))

    ;; now should be looking at:
    ;; ?\x10		type code 16 is author/work description
    ;; ?\x00		level = 0 (author)
    ;; [1 byte]		length of the following description
    ;; [n bytes]	author description

    ;; just make sure
    (if (or (/= (following-char) ?\x10)
	    (/= (char-after (1+ (point))) ?\x00))
      (error "Unexpected author description introducer %02x%02x"
	     (following-char)
	     (char-after (1+ (point)))))

    (forward-char 3)
    (setq desc (cgreek-tlg-read-string (preceding-char)))

    ;; print out author description
    ;; author id value is given as property
    (save-excursion
      (set-buffer cooked)
      (if (/= (point) (point-min))
	  (insert "\n"))
      (add-text-properties
       (point)
       (progn (insert (format "%s\n" desc)) (point))
       (list 'tlg-author-id id)))

    ;; process each work written by this author
    (while (= (following-char) cgreek-tlg-new-work)
      (cgreek-tlg-parse-work cooked))

    (goto-char next-author)))

(defun cgreek-tlg-parse-work (cooked)
  "Process a work entry in TLG IDT file."

  (let (id desc next-work block cit-desc plist)

    ;; should be looking at:
    ;; ?\x02		type code 2 marks new work
    ;; [2 bytes]	length of the work section in bytes
    ;; [2 bytes]	work location in blocks
    ;; ?\xef		left nibble = 1110 : escape code
    ;;			right nibble = 1111 : ASCII string
    ;; ?\x81		escape level = 1 (work)
    ;; [n bytes]	ID value in string (with MSB on)
    ;; ?\xff		ASCII string terminator

    ;; just make sure
    (if (or (/= (char-after (+ (point) 5)) ?\xef)
	    (/= (char-after (+ (point) 6)) ?\x81))
	(error "Unexpected work ID value introducer %02x%02x"
	       (char-after (+ (point) 5))
	       (char-after (+ (point) 6))))
    
    (setq next-work
	  (+ (point)
	     (* (char-after (1+ (point))) 256)
	     (char-after (+ (point) 2))))
    (setq block
	  (+ (* (char-after (+ (point) 3)) 256)
	     (char-after (+ (point) 4))))
    (forward-char 7)
    (setq id (cgreek-tlg-read-string))

    ;; now should be looking at:
    ;; ?\x10		type code 16 is author/work description
    ;; ?\x01		level = 1 (work)
    ;; [1 byte]		length of the following description
    ;; [n bytes]	work description

    ;; just make sure
    (if (or (/= (following-char) ?\x10)
	    (/= (char-after (1+ (point))) ?\x01))
	(error "Unexpected work description introducer %02x%02x"
	       (following-char)
	       (char-after (1+ (point)))))

    (forward-char 3)
    (setq desc (cgreek-tlg-read-string (preceding-char)))
    
    (if (= (following-char) cgreek-tlg-citation)

	;; citation descriptions are given for this work; set them
	(while (= (following-char) cgreek-tlg-citation)
	  (setq cit-desc (append (cgreek-tlg-parse-citation) cit-desc)))

      ;; citation descriptions are not given for this work;
      ;; inherit previous work's ones
      (save-excursion
	(set-buffer cooked)
	(save-excursion
	  (forward-line -1)
	  (setq cit-desc
		(get-text-property (point) 'tlg-citation-description)))))

    ;; set property list
    (setq plist
	  (list 'tlg-work-id id
		'tlg-citation-description cit-desc
		'tlg-starting-block block
		'mouse-face 'highlight))

    ;; special consideration for Bekker pages
    ;; Bekker page is, if present, always at level y (level 5)
    (when (string= (plist-get cit-desc 5) "Bekker page")
      (setq plist
	    (append
	     (list 'tlg-first-bekker-page
		   (cgreek-tlg-get-bekker-page))
	     plist)))

    ;; print out work description
    ;; necessary information is given as properties
    (save-excursion
      (set-buffer cooked)
      (add-text-properties
       (point)
       (progn (insert (format "\t%s" desc)) (point))
       plist)
      (insert "\n"))

    ;; jump to the starting point of the next work and return
    (goto-char next-work)))

(defun cgreek-tlg-parse-citation ()
  "Process a citation description entry in TLG IDT file.
Return a list of the form (LEVEL DESCRIPTION).
LEVEL is an integer varying from 1 (level v) to 5 (level z, or n).
DESCRIPTION is a string."

  (let (level length)

  ;; should be looking at cgreek-tlg-citation, followed by 1-byte
  ;; level, followed by 1-byte length
  ;; ?\x11		type code 17 is citation description
  ;; [1 byte]		level (0 = n or z, 1 = y, 2 = x, 3 = w, 4 = v)
  ;; [1 byte]		length of the following description
  ;; [n bytes]		citation description
  
    (setq level (- 6 (char-after (1+ (point)))) ; 2 = v, ... , 6 = z or n
	  length (char-after (+ (point) 2)))
    (list level
	  (buffer-substring
	   (+ (point) 3)
	   (progn (goto-char (+ (point) 3 length)) (point))))))

(defun cgreek-tlg-read-string (&optional length)
  "Return TLG string starting at point.  Point is moved.

Without the optional argument LENGTH, read up to \xff, stripping each MSB.

With the optional argument LENGTH, read that long beta code and return \0.
The beta code read will be stored in cgreek-tlg-temporary-buffer
for later conversion.
Each beta code fragment is separated by \0."

  (if length

      (let ((str (buffer-substring (point) (+ (point) length))))
	(forward-char length)
	(save-excursion
	  (set-buffer cgreek-tlg-temporary-buffer)

	  ;; work around for the error in Disk D
;	  (if (string= cgreek-tlg-cdrom-version "D")
;	      (insert "&"))
	  (insert "&")

	  (insert str 0))
	(char-to-string 0))

    (let ((p (point))
	  (i 0)
	  str)
      (skip-chars-forward "^\377")
      (forward-char 1)
      (setq str (buffer-substring p (1- (point))))
      (while (< i (length str))
	(aset str i (- (aref str i) 128))
	(setq i (1+ i)))
      str)))

(defun cgreek-tlg-parse-txt-mouse (click)
  "Invoke cgreek-tlg-parse-txt with parameters given to the clicked line."
  (interactive "e")
  (let* ((start (event-start click))
	 (window (car start))
	 (pos (car (cdr start))))
    (select-window window)
    (goto-char pos))
  (cgreek-tlg-parse-txt-keyboard))

(defun cgreek-tlg-parse-txt-keyboard ()
  "Invoke cgreek-tlg-parse-txt with parameters given to the current line."
  (interactive)
  (if (get-text-property (point) 'tlg-work-id) ; on a valid line?

      (let (author-id author-name (cit (current-buffer)))

	;; get author-id and author-name given as text properties
	;; work name in a line is preceded by a tab; author name is not
	(save-excursion
	  (re-search-backward "^[^\t]")	; should be found
	  (setq author-id (get-text-property (point) 'tlg-author-id))
	  (setq author-name
		(buffer-substring-no-properties
		 (progn (beginning-of-line) (point))
		 (progn (end-of-line) (point)))))

	;; parse and display
	(switch-to-buffer
	 (cgreek-tlg-parse-txt
	  (buffer-name)
	  author-id
	  author-name
	  (get-text-property (point) 'tlg-work-id)
	  (save-excursion
	    (buffer-substring-no-properties ; this is work-name
	     (progn (beginning-of-line)
		    (1+ (point)))	; skipping line-initail tab
	     (progn (end-of-line) (point))))
	  (get-text-property (point) 'tlg-citation-description)
	  (get-text-property (point) 'tlg-starting-block))))))

(defun cgreek-tlg-parse-txt (idt-buffer author-id author-name work-id work-name
				     cit-desc &optional block)
  "Parse the work WORK-ID of the author AUTHOR-ID from a tlg txt file.

The tlg txt file name \"/.../tlg????.txt\" is derived from IDT-BUFFER,
which is of the form \"*tlg????.idt*\".

Both AUTHOR-ID and WORK-ID are strings although they consist of digits.

AUTHOR-NAME and WORK-NAME are descriptive string.

CIT-DESC is a plist of the form (level-1 desc-1 level-2 desc-2 ...) where
level-x is an integer from 1 (level v) to 5 (level z or n) and
desc-x is a string like \"line\", \"section\", etc.

BLOCK is an integer representing the starting block of WORK-ID.

Return the buffer in which parsed result is contained."

  (let ((txt (get-buffer work-name))
	(cit (get-buffer (concat work-name " citation"))))

    ;; Am I already created?
    (if (and txt
	     cit
	     (save-excursion
	       (set-buffer txt)
	       (and
		(string= author-name cgreek-tlg-author-name)
		(string= work-name cgreek-tlg-work-name))))

	;; OK, already created
	txt

      ;; no, not created yet
      (message "Reading %s ..." work-name)
      (setq txt (generate-new-buffer work-name))
      (setq cit (generate-new-buffer (concat work-name " citation")))

      ;; cgreek-tlg2citation-program writes citation information to stdout
      ;; and beta code text to stderr
      (call-process
       cgreek-tlg2citation-program
       (expand-file-name
	(concat (substring idt-buffer 1 (- (length idt-buffer) 4))
		"txt")		   ; *tlg????.idt* -> /.../tlg????.txt
	cgreek-tlg-directory)
       (list cit cgreek-tlg-temporary-file)
       nil
       "-a" author-id "-w" work-id "-s" (format "%d" block))

      ;; beta code text just written is decoded and inserted into txt
      (call-process
       cgreek-tlg2emacs-program
       cgreek-tlg-temporary-file
       txt)
      (set-buffer txt)

      ;; buffer local bindings
      (setq cgreek-tlg-author-name author-name
	    cgreek-tlg-work-name work-name
	    cgreek-tlg-citation-buffer cit
	    cgreek-tlg-citation-description cit-desc
	    cgreek-tlg-idt-buffer idt-buffer)
      (local-set-key "q" '(lambda () (interactive) (kill-buffer nil)))
      (local-set-key "a" 'cgreek-tlg-parse-authtab)
      (local-set-key "w" 'cgreek-tlg-parse-idt-keyboard)
      (local-set-key "j" 'cgreek-tlg-jump)
      ;; This is a workaround for a bug in Emacs-21.3
      (local-set-key "\C-l"
		     '(lambda ()
			(interactive)
			(set-buffer-multibyte nil)
			(set-buffer-multibyte t)
			(recenter)))
      (add-hook 'post-command-hook
		'cgreek-tlg-show-citation
		nil t)
      (add-hook 'kill-buffer-hook
		'(lambda ()
		   (if (buffer-live-p cgreek-tlg-citation-buffer)
		       (kill-buffer cgreek-tlg-citation-buffer))
		   (let ((win (get-buffer-window "*TLG citation*" t)))
		     (if (window-live-p win)
			 (save-selected-window
			   (select-window win)
			   (if (not (one-window-p t))
			       (delete-window win)))))
		   (if (buffer-live-p (get-buffer "*TLG citation*"))
		       (kill-buffer "*TLG citation*")))
		nil t)

      ;; It is better to remove the final newline so that the cursor
      ;; always stays on a `real' line.
      (goto-char (point-max))
      (if (= (preceding-char) ?\n)
	  (delete-backward-char 1))

      (goto-char (point-min))
      (set-buffer-modified-p nil)
      (setq buffer-read-only t)
      (message nil)

      ;; return value
      txt)))

;; why is this not supplied by default?
(defun current-line ()
  "Return current line, beginning with 1 (like in goto-line)."
  (+ (count-lines (point-min) (point))
     (if (= (current-column) 0) 1 0)))

(defun cgreek-tlg-show-citation ()
  "Display current buffer's tlg citation in buffer \"*TLG citation*\""

  (let ((author cgreek-tlg-author-name)
	(work cgreek-tlg-work-name)
	(cit-desc cgreek-tlg-citation-description)
	(mesg nil)
	(line (current-line)))

    ;; AUTHOR should be non-nil if called from a tlg txt buffer
    (when author

      (save-excursion
	(set-buffer cgreek-tlg-citation-buffer) ; current buffer's citation
	(goto-line line)
	(beginning-of-line)

	(when (looking-at cgreek-tlg-citation-regexp)

	  ;; (match-string 1) = level n citation ID value
	  ;; (match-string 2) = level v citation ID value
	  ;; (match-string 3) = level w citation ID value
	  ;; (match-string 4) = level x citation ID value
	  ;; (match-string 5) = level y citation ID value
	  ;; (match-string 6) = level z citation ID value

	  (if (or (string= (match-string 1) "")
		  (null (match-string 1)))

	      ;; no level n, thus this work is hierarchical.
	      ;; level z is always present.  level v..y may or may not.
	      (setq mesg
		    (mapcar
		     '(lambda (match)
			(if (or (string= (match-string match) "")
				(null (match-string match)))
			    ""
			  (format "%s %s   "
				  (plist-get cit-desc match)
				  (match-string match))))
		     '(2 3 4 5 6)))

	    ;; level n is present, thus this work is a document.
	    ;; level v..z may exist, buf their description is not available.
	    (setq mesg
		  (mapcar
		   '(lambda (match)
		      (if (or (string= (match-string match) "")
			      (null (match-string match)))
			  ""	
			(format "%s   " (match-string match))))
		   '(2 3 4 5 6)))

	    ;; only level n has description, but it locates at level 6!
	    (setq mesg
		  (cons (format "%s "
				(plist-get cit-desc 6))
			(cons (format "%s   " (match-string 1))
			      mesg))))

	  ;; display citation
	  (set-buffer (get-buffer-create "*TLG citation*"))
	  (erase-buffer)
	  (setq truncate-lines t)
	  (insert author "\n" work "\n" (apply 'concat mesg))
	  (unless (get-buffer-window (current-buffer) 'visible)
	    (switch-to-buffer-other-window (current-buffer))
	    (shrink-window-if-larger-than-buffer (selected-window)))))

      ;; put cursor back to the txt buffer
      (select-window (get-buffer-window (current-buffer))))))

(defun cgreek-tlg-jump ()
  "Jump to the specified position in the current tlg txt buffer"
  (interactive)
  (let ((txt-buf (current-buffer))
	(cit-buf cgreek-tlg-citation-buffer)
	(cit-desc cgreek-tlg-citation-description)
	level value matched desc line)

    ;; this function is meaningful only in a tlg txt buffer,
    ;; in that case cit-desc should be non-nil
    (when cit-desc

      (set-buffer cit-buf)
      (goto-char (point-min))
      (looking-at cgreek-tlg-citation-regexp)
      (if (string= (match-string 1) "")	; 1 means level n

	  ;; level n id does not exist, that means this is a hierarchical text
	  (progn

	    (setq level 2)		; 2 means level v
	    (while (<= level 6)		; 6 means level z
	      (setq desc (plist-get cit-desc level))

	      ;; prompt only for levels with description
	      (when desc

		;; read id
		(setq value (read-from-minibuffer (concat desc ": ")))
		(if (string= value "")
		    (error ""))

		;; special handling for Bekker page
		(when (string= desc "Bekker page")
		  (cgreek-tlg-open-bekker-page value)
		  (setq txt-buf (current-buffer)
			cit-buf cgreek-tlg-citation-buffer
			cit-desc cgreek-tlg-citation-description)
		  (set-buffer cit-buf)
		  (goto-char (point-min)))

		;; start searching
		(setq matched nil)
		(while (not matched)
		  (looking-at cgreek-tlg-citation-regexp)
		  (if (string= (match-string level) value)
		      (setq matched t)
		    (if (/= (forward-line 1) 0)
			(setq matched 'eob))))

		;; really found?
		(if (eq matched 'eob)
		    (error "Invalid value for this work")

		  (setq line (current-line))
		  (switch-to-buffer txt-buf) ; show it to the user
		  (goto-line line)
		  (recenter 0)
		  (cgreek-tlg-show-citation)
		  (set-buffer cit-buf)))

	      (setq level (1+ level))))

	;; level n id exist, thus this is a `document' (non-hierarchical)
	(setq value
	      (read-from-minibuffer
	       ;; level n DESCRIPTION is tagged as 6, but ... (see below)
	       (concat (plist-get cit-desc 6) ": ")))
	(if (string= value "")
	    (error ""))

	;; start searching
	(setq matched nil)
	(while (not matched)
	  (looking-at cgreek-tlg-citation-regexp)
	  (if (string= (match-string 1) value) ; level n CITATION id is at 1!
	      (setq matched t)
	    (if (/= (forward-line 1) 0)
		(setq matched 'eob))))

	;; really found?
	(if (eq matched 'eob)
	    (error "Invalid value for this work")

	  (setq line (current-line))
	  (switch-to-buffer txt-buf)	; show it to the user
	  (goto-line line)
	  (recenter 0)
	  (cgreek-tlg-show-citation))))))

(defun cgreek-tlg-open-bekker-page (page)
  "Open the work in which Bekker page PAGE resides."
  (let ((distance 9999)
	candidate
	firstpage)
    (set-buffer (cgreek-tlg-parse-idt "tlg0086.idt")) ; Aristotle's work list
    (goto-char (point-min))
    (while (not (eobp))
      (if (and (setq firstpage
		     (plist-get (text-properties-at (point))
				'tlg-first-bekker-page))
	       (cgreek-tlg-bekker-<= firstpage page)
	       (< (cgreek-tlg-bekker-distance page firstpage) distance))
	  (setq candidate (current-line)
		distance (cgreek-tlg-bekker-distance page firstpage)))
      (beginning-of-line 2))		; next line's beginning
    (goto-line candidate)
    (cgreek-tlg-parse-txt-keyboard)))

(defun cgreek-tlg-get-bekker-page ()
  "Return the Bekker page number in a starting citation ID"

  ;; should be looking at:
  ;; ?\x03		type code 3 marks new section
  ;; [2 bytes]		the text block in which this section is located
  ;; ?\x08		type code 8 marks the section starting ID

  ;; just make sure
  (if (or (/= (following-char) ?\x03)
	  (/= (char-after (+ (point) 3)) ?\x08))
      (error "Unexpected subdivision %02x, %02x"
	     (following-char)
	     (char-after (+ (point) 3))))

  (forward-char 4)
  (if (= (following-char) ?\xa1)	; this happens only in two works
      (forward-char 1))

  ;; in TLG CD-ROM volume E, there are only two cases listed below
  (cond

   ;; left nibble = 1001 : level y
   ;; right nibble = 1001 : 7-bit numeric value + single ASCII character
   ((= (following-char) ?\x99)
    (format "%d%c"
	    (logand (char-after (1+ (point))) ?\x7f)
	    (logand (char-after (+ (point) 2)) ?\x7f)))

   ;; left nibble = 1001 : level y
   ;; right nibble = 1100 : 14-bit numeric value + single ASCII character
   ((= (following-char) ?\x9c)
    (format "%d%c"
	    (+ (* (logand (char-after (1+ (point))) ?\x7f) 128)
	       (logand (char-after (+ (point) 2)) ?\x7f))
	    (logand (char-after (+ (point) 3)) ?\x7f)))

   ;; should never happen
   (t
    (error "Unexpected starting citation code %02x" (following-char)))))

(defun cgreek-tlg-bekker-<= (x y)
  "Return t if Bekker page X is `less than' or `equal to' Bekker page Y."
  (let ((x-num (string-to-number x))
	(y-num (string-to-number y)))
    (cond
     ((< x-num y-num) t)
     ((> x-num y-num) nil)
     ((or (= (aref x (1- (length x))) ?a) 
	  (= (aref y (1- (length y))) ?b))
      t)
     (t nil))))
    
(defun cgreek-tlg-bekker-distance (x y)
  "Return Bekker page X `minus' Bekker page Y.  Both X and Y are strings.
A Bekker page consists of a numeric part and a character, either a or b.
The characters `a' and `b' are regarded as .0 and .1, respectively.
For example, 45a - 23b = 45.0 - 23.1 = 21.9."
  (- (+ (string-to-number x) (if (= (aref x (1- (length x))) ?a) 0.0 0.1))
     (+ (string-to-number y) (if (= (aref y (1- (length y))) ?a) 0.0 0.1))))

(provide 'cgreek-tlg)

;;; cgreek-tlg.el ends here
