/*
tlg2cit.c --- extract citation information from a TLG TXT file.

Copyright (C) 2003, 2005
National Institute of Advanced Industrial Science and Technology (AIST)
Registration Number: H15PRO 109

Author: TAKAHASHI Naoto <ntakahas@m17n.org>
Keywords: multilingual, classical Greek, Unicode, LaTeX, Ibycus4

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.
*/

/* description */

/*
Structure of a TLG TXT file
---------------------------

FILE := [BLOCK ...] LAST_BLOCK
BLOCK := RECORD [RECORD ...] EOB [PAD ...]
LAST_BLOCK := RECORD [RECORD ...] EOF EOB [PAD ...]
RECORD := ID_BYTES TEXT
ID_BYTES := MSB [MSB ...]
TEXT := ASCII [ASCII ...]
MSB := '0x80' | '0x81' |  ...  | '0xff'
ASCII := ' ' | '!' |  ...  | '~'
EOB := '0xfe'
EOF := '0xf0'
PAD := '\0'

Each block (BLOCK and LAST_BLOCK) contains 8192 bytes.
Null ('\0') is used for padding.


Structure of ID bytes
---------------------

1. When the first byte is between \x80 and \xdf inclusive

The left nibble represents the citation level to which ID value is
assigned.

1000 ????  ==  \x8?  :  level z
1001 ????  ==  \x9?  :  level y
1010 ????  ==  \xa?  :  level x
1011 ????  ==  \xb?  :  level w
1100 ????  ==  \xc?  :  level v
1101 ????  ==  \xd?  :  level n

The right nibble represents the value that is set to the specified
citation level.

1??? 0000  ==  \x?0  :  increment current numeric value
1??? 0001  ==  \x?1  :  numeric value 1
1??? 0010  ==  \x?2  :  numeric value 2
1??? 0011  ==  \x?3  :  numeric value 3
1??? 0100  ==  \x?4  :  numeric value 4
1??? 0101  ==  \x?5  :  numeric value 5
1??? 0110  ==  \x?6  :  numeric value 6
1??? 0111  ==  \x?7  :  numeric value 7

If the right nibble is greater than 7, it is the following byte(s)
that represent the value set to the specified citation level.  You
need to strip the sign bit to interpret the value.

1??? 1000  ==  \x?8  :  7-bit numeric value
1??? 1001  ==  \x?9  :  7-bit numeric value + 7-bit ASCII char
1??? 1010  ==  \x?a  :  7-bit numeric value + ASCII string
1??? 1011  ==  \x?b  :  14-bit numeric value
1??? 1100  ==  \x?c  :  14-bit numeric value + 7-bit ASCII char
1??? 1101  ==  \x?d  :  14-bit numeric value + ASCII string
1??? 1110  ==  \x?e  :  current numeric value + new 7-bit ASCII char
1??? 1111  ==  \x?f  :  ASCII string (no numeric value)

A string is terminated by a \xff.

2. When the first byte is between \xe0 and \xef inclusive

The citation level is specified by the second byte.

\xe?  \x80  :  level a
\xe?  \x81  :  level b
\xe?  \x82  :  level c
\xe?  \x83  :  level d

(The second byte may be between \xe0 and \xfa inclusive.
In this case it specifies descriptor ID, which we skip here.)

If the right nibble of the first byte are less than 8, it
represents the value set to the specified citation level just like
above.  Otherwise, the following byte(s) do so like above, but in
this case "the following byte(s)" means "the third byte" or "the
sequence of bytes beginning at the third byte" since the second
byte is used to specify the citation level.

3. When the first byte is greater than or equal to \xf0

\xff  :  end of ASCII string
\xfe  :  end of block
\xf0  :  end of file
\xf8  :  exception start
\xf9  :  exception end

\xff terminates the string value.

\xfe preceded by zero or one \xf0 and followed by zero or more \x0
(or "\xf0?\xfe\x0*" if you prefer regular expression) are used for
padding to fit 8K boundary.  They can be ignored safely.

\xf8 and \xf9 can be ignored safely, too.
*/

/* code */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* macros for citation levels */
#define LEVEL_A 0		/* author */
#define LEVEL_B 1		/* work */
#define LEVEL_C 2		/* abbr. work */
#define LEVEL_D 3		/* abbr. author  */
#define LEVEL_N 4		/* if set, Z to V are unhierarchical */
#define LEVEL_V 5		/* sth higher than W (volume, etc) */
#define LEVEL_W 6		/* sth higher than X (chapter, etc) */
#define LEVEL_X 7		/* sth higher than Y (section, etc) */
#define LEVEL_Y 8		/* sth higher than Z (paragraph, etc) */
#define LEVEL_Z 9		/* line number */

/*
Considering citations like "1a", we save every citation ID value,
even a numeric value, in the form of string.  The first index of
citation[][] corresponds to citation level.
*/
#define BUFSIZE 256
char citation[10][BUFSIZE];

/* garbage can */
char devnull[256];

/*
Remember, ID bytes range from \x80 to \xff.  Thus the sign bit must
be stripped to interpret a numeric value and an ASCII string value.
*/

int read_7bit(FILE *f)
     /* f: input stream */
{
  return(getc(f) & 0x7f);
}

int read_14bit(FILE *f)
     /* f: input stream */
{
  int i;

  i = (getc(f) & 0x7f) << 7;
  return(i + (getc(f) & 0x7f));
}

void read_string(FILE *f, int bsize, char *ptr)
     /*
      * f: input stream
      * bsize: buffer size of ptr
      * ptr: where the read string is saved
      */
{
  int i;

  /* 0xff (not \0) terminates input string */
  while ((bsize > 1) && ((i = getc(f)) != 0xff)) {
    bsize--;
    *ptr++ = (char)(i & 0x7f);
  }
  *ptr = '\0';
}

void set_value(FILE *f, char *ptr, int op)
     /*
      * f: input stream
      * ptr: where ID value is saved
      * op: the right nibble
      *
      * Read citation ID value from stdin and save it as string to the
      * place pointed by ptr.  The third argument op specifies ID
      * value type.
      */
{
  int i, used;
  char *last;

  switch (op) {

  case 0:
    /* increment the ID at this level */
    last = ptr + strlen(ptr) - 1;

    if (isdigit(*last)) {
      /* there is a possibility that "1col1" is incremented to "1col2" */
      while (isdigit(*(last - 1))) {
	last--;
      }
      i = atoi(last);
      snprintf(last, BUFSIZE - (ptr - last), "%d", i + 1);
    }

    else {
      /* ending with a character, e.g. "10a" --> "10b" */
      (*last)++;
    }
    break;

  case 1: case 2: case 3: case 4: case 5: case 6: case 7:
    /* literal binary ID value */
    snprintf(ptr, BUFSIZE, "%d", op);
    break;

  case 8:
    /* 7-bit binary value */
    snprintf(ptr, BUFSIZE, "%d", read_7bit(f));
    break;

  case 9:
    /* 7-bit binary value + single ASCII char */
    i = read_7bit(f);
    snprintf(ptr, BUFSIZE, "%d%c", i, (char)getc(f) & 0x7f);
    break;

  case 10:
    /* 7-bit binary value + ASCII string */
    used = snprintf(ptr, BUFSIZE, "%d", read_7bit(f));
    read_string(f, BUFSIZE - used, ptr + used);
    break;

  case 11:
    /* 14-bit binary value */
    snprintf(ptr, BUFSIZE, "%d", read_14bit(f));
    break;

  case 12:
    /* 14-bit binary value + single ASCII char */
    i = read_14bit(f);
    snprintf(ptr, BUFSIZE, "%d%c", i, (char)getc(f) & 0x7f);
    break;

  case 13:
    /* 14-bit binary value + ASCII string */
    used = snprintf(ptr, BUFSIZE, "%d", read_14bit(f));
    read_string(f, BUFSIZE - used, ptr + used);
    break;

  case 14:
    /* same binary value + new single ASCII char */
    i = atoi(ptr);
    snprintf(ptr, BUFSIZE, "%d%c", i, (char)getc(f));
    break;

  case 15:
    /* no binary value + ASCII string */
    read_string(f, BUFSIZE, ptr);
    break;

  default:
    /* should never happen */
    fprintf(stderr, "illegal operand %d in set_value\n", op);
    exit(-1);
  }
}

/* whether seeing requested work or not */
#define STANDBY 0
#define ACTIVE 1

static char help_string[] =
"Usage: %s -a author_id -w work_id [-s starting_block] [input_file]\n";

main(int argc, char *argv[])
{
  int i, byte, level, op;
  FILE *f = stdin;
  int start = 0;		/* starting block */
  char *author = NULL;		/* author-id as string */
  char *work = NULL;		/* work-id as string */
  int status = STANDBY;

  /* argument analysis */
  i = 1;
  while (i < argc) {
    if ((strcmp(argv[i], "-s") == 0) ||
	(strcmp(argv[i], "--starting_block") == 0)) {
      start = atoi(argv[i + 1]);
      i += 2;
    }
    else if ((strcmp(argv[i], "-a") == 0) ||
	     (strcmp(argv[i], "--author_id") == 0)) {
      author = argv[i + 1];
      i += 2;
    }
    else if ((strcmp(argv[i], "-w") == 0) ||
	     (strcmp(argv[i], "--work_id") == 0)) {
      work = argv[i + 1];
      i += 2;
    }
    else if ((strcmp(argv[i], "-h") == 0) ||
	     (strcmp(argv[i], "--help") == 0)) {
      fprintf(stderr, help_string, argv[0]);
      exit(0);
    }
    else {
      f = fopen(argv[i], "r");
      if (!f) {
	fprintf(stderr, "cannot open %s\n", argv[i]);
	exit(-1);
      }
      fseek(f, start * 8192, SEEK_SET);
      i++;
    }
  }

  if ((author == NULL) || (work == NULL)) {
    fprintf(stderr, help_string, argv[0]);
    exit(-1);
  }

  /* initialise ID values */
  for (i = 0; i < 10; i++) {
    citation[i][0] = '\0';
  }

  while ((byte = getc(f)) != EOF) {

    if ((byte >= 0x20) && (byte <= 0x7e)) {
      /* text in beta code, which means ID bytes have just ended */

      /* status check */
      if ((status == STANDBY) &&
	  (strcmp(&citation[LEVEL_A][0], author) == 0) &&
	  (strcmp(&citation[LEVEL_B][0], work) == 0)) {
	status = ACTIVE;
      }
      if ((status == ACTIVE) && (strcmp(&citation[LEVEL_B][0], work) != 0)) {
	exit(0);		/* finished */
      }

      /* print the gathered citation data in tab-separated form */
      if (status == ACTIVE) {
	for (i = LEVEL_N; i <= LEVEL_Z; i++) {
	  printf("%s\t", &citation[i][0]);
	}
	printf("\n");
      }

      /* output beta code to stderr only in active status */
      do {
	if (status == ACTIVE) {
	  fprintf(stderr, "%c", byte);
	}
	byte = getc(f);
      } while ((byte >= 0x20) && (byte <= 0x7e));

      /* beta code lines will be separeted by \0, not \n */
      if (status == ACTIVE) {
	fprintf(stderr, "%c", 0);
      }

      if (byte == EOF) {
	/* should never happen */
	fprintf(stderr, "unexpected EOF\n");
	exit(-1);
      }
    }

    if ((byte >= 0x80) && (byte <= 0xdf)) {
      /*
       * left nibble -> citation level (from z to n; 1000 = z, 1001 = y,
       *                1010 = x, 1011 = w, 1100 = v, 1101 = n)
       * right nibble -> operation
       */

      level = 9 - ((byte & 0x70) >> 4);	/* now n = 4, v = 5, ... z = 9 */
      op = byte & 0x0f;
      set_value(f, &(citation[level][0]), op);

      if (level != LEVEL_N) {
	/* reset all lower levels */
	for (i = level + 1; i <= LEVEL_Z; i++) {
	  if (citation[level][0] == 't') {
	    citation[i][0] = '\0';
	  }
	  else {
	    citation[i][0] = '1';
	    citation[i][1] = '\0';
	  }
	}
      }
    }
    
    else if ((byte >= 0xe0) && (byte <= 0xef)) {
      /*
       * right nibble of this byte -> operation
       * next byte -> citation level
       */

      op = byte & 0x0f;
      byte = getc(f);
      if (byte == EOF) {
	/* should never happen */
	fprintf(stderr, "unexpected EOF\n");
	exit(-1);
      }
      level = byte & 0x7f;

      if (level <= 3) {
	/* citation level a to d (a = 0, b = 1, c = 2, d = 3) */
	set_value(f, &(citation[level][0]), op);
	/* reset all lower levels to null */
	for (i = level + 1; i <= LEVEL_Z; i++) {
	  citation[i][0] = '\0';
	}
      }
      else if ((level >= 97) && (level <= 122)){
	/* not a citation level but a descriptor; skip */
	set_value(f, devnull, op);
      }
      else {
	/* should never happen */
	fprintf(stderr, "invalid escape %d\n", level);
	exit(-1);
      }
    }

    else if ((byte == 0xfe) || (byte == 0xf0) || (byte == 0) ||
	     (byte == 0xf8) || (byte == 0xf9)) {
      /* just ignore them */
    }

    else {
      /* should never happen */
      fprintf(stderr, "unexpected code %x\n", byte);
      exit(-1);
    }
  }
}
